#############################################################################
# 				Documentation				    #
#############################################################################

# Author:   Todd Whiteman
# Date:     16th March, 2009
# Verion:   2.0.0
# License:  Public Domain - free to do as you wish
# Homepage: http://twhiteman.netfirms.com/des.html
#
# This is a pure python implementation of the DES encryption algorithm.
# It's pure python to avoid portability issues, since most DES 
# implementations are programmed in C (for performance reasons).
#
# Triple DES class is also implemented, utilising the DES base. Triple DES
# is either DES-EDE3 with a 24 byte key, or DES-EDE2 with a 16 byte key.
#
# See the README.txt that should come with this python module for the
# implementation methods used.
#
# Thanks to:
#  * David Broadwell for ideas, comments and suggestions.
#  * Mario Wolff for pointing out and debugging some triple des CBC errors.
#  * Santiago Palladino for providing the PKCS5 padding technique.
#  * Shaya for correcting the PAD_PKCS5 triple des CBC errors.
#
"""A pure python implementation of the DES and TRIPLE DES encryption algorithms.

Class initialization
--------------------
pyDes.des(key, [mode], [IV], [pad], [padmode])
pyDes.triple_des(key, [mode], [IV], [pad], [padmode])

key     -> Bytes containing the encryption key. 8 bytes for DES, 16 or 24 bytes
	   for Triple DES
mode    -> Optional argument for encryption type, can be either
	   pyDes.ECB (Electronic Code Book) or pyDes.CBC (Cypher Block Chaining)
IV      -> Optional Initial Value bytes, must be supplied if using CBC mode.
	   Length must be 8 bytes.
pad     -> Optional argument, set the pad character (PAD_NORMAL) to use during
	   all encrypt/decrpt operations done with this instance.
padmode -> Optional argument, set the padding mode (PAD_NORMAL or PAD_PKCS5)
	   to use during all encrypt/decrpt operations done with this instance.

I recommend to use PAD_PKCS5 padding, as then you never need to worry about any
padding issues, as the padding can be removed unambiguously upon decrypting
data that was encrypted using PAD_PKCS5 padmode.

Common methods
--------------
encrypt(data, [pad], [padmode])
decrypt(data, [pad], [padmode])

data    -> Bytes to be encrypted/decrypted
pad     -> Optional argument. Only when using padmode of PAD_NORMAL. For
	   encryption, adds this characters to the end of the data block when
	   data is not a multiple of 8 bytes. For decryption, will remove the
	   trailing characters that match this pad character from the last 8
	   bytes of the unencrypted data block.
padmode -> Optional argument, set the padding mode, must be one of PAD_NORMAL
	   or PAD_PKCS5). Defaults to PAD_NORMAL.
	  


Note: This code was not written for high-end systems needing a fast
      implementation, but rather a handy portable solution with small usage.

"""

import sys
import string, re

# _pythonMajorVersion is used to handle Python2 and Python3 differences.
_pythonMajorVersion = sys.version_info[0]

# Modes of crypting / cyphering
ECB =	0
CBC =	1

# Modes of padding
PAD_NORMAL = 1
PAD_PKCS5 = 2

# PAD_PKCS5: is a method that will unambiguously remove all padding
#            characters after decryption, when originally encrypted with
#            this padding mode.
# For a good description of the PKCS5 padding technique, see:
# http://www.faqs.org/rfcs/rfc1423.html

# The base class shared by des and triple des.
class _baseDes(object):
	def __init__(self, mode=ECB, IV=None, pad=None, padmode=PAD_NORMAL):
		#if IV:
			#IV = self._guardAgainstUnicode(IV)
		#if pad:
			#pad = self._guardAgainstUnicode(pad)
		self.block_size = 64
		# Sanity checking of arguments.
		if pad and padmode == PAD_PKCS5:
			raise ValueError("Cannot use a pad character with PAD_PKCS5")
		if IV and len(IV) != self.block_size:
			raise ValueError("Invalid Initial Value (IV), must be a multiple of " + str(self.block_size) + " bits")

		# Set the passed in variables
		self._mode = mode
		self._iv = IV
		self._padding = pad
		self._padmode = padmode

	def getKey(self):
		"""getKey() -> bytes"""
		return self.__key

	def setKey(self, key):
		"""Will set the crypting key for this object."""
		#key = self._guardAgainstUnicode(key)
		self.__key = key

	def getMode(self):
		"""getMode() -> pyDes.ECB or pyDes.CBC"""
		return self._mode

	def setMode(self, mode):
		"""Sets the type of crypting mode, pyDes.ECB or pyDes.CBC"""
		self._mode = mode

	def getPadding(self):
		"""getPadding() -> bytes of length 1. Padding character."""
		return self._padding

	def setPadding(self, pad):
		"""setPadding() -> bytes of length 1. Padding character."""
		if pad is not None:
			pad = self._guardAgainstUnicode(pad)
		self._padding = pad

	def getPadMode(self):
		"""getPadMode() -> pyDes.PAD_NORMAL or pyDes.PAD_PKCS5"""
		return self._padmode
		
	def setPadMode(self, mode):
		"""Sets the type of padding mode, pyDes.PAD_NORMAL or pyDes.PAD_PKCS5"""
		self._padmode = mode

	def getIV(self):
		"""getIV() -> bytes"""
		return self._iv

	def setIV(self, IV):
		"""Will set the Initial Value, used in conjunction with CBC mode"""
		if not IV or len(IV) != self.block_size:
			raise ValueError("Invalid Initial Value (IV), must be a multiple of " + str(self.block_size) + " bytes")
		IV = self._guardAgainstUnicode(IV)
		self._iv = IV

	def _padData(self, data, pad, padmode):
		# Pad data depending on the mode
		if padmode is None:
			# Get the default padding mode.
			padmode = self.getPadMode()
		if pad and padmode == PAD_PKCS5:
			raise ValueError("Cannot use a pad character with PAD_PKCS5")

		if padmode == PAD_NORMAL:
			if len(data) % self.block_size == 0:
				# No padding required.
				return data

			if not pad:
				# Get the default padding.
				pad = self.getPadding()
			if not pad:
				raise ValueError("Data must be a multiple of " + str(self.block_size) + " bits in length. Use padmode=PAD_PKCS5 or set the pad character.")
			data += (self.block_size - (len(data) % self.block_size)) * pad
		
		elif padmode == PAD_PKCS5:
			pad_len = 64 - (len(data) % self.block_size)
			if _pythonMajorVersion < 3:
				data += pad_len * chr(pad_len)
			else:
				data += bytes([pad_len] * pad_len)

		return data

	def _unpadData(self, data, pad, padmode):
		# Unpad data depending on the mode.
		if not data:
			return data
		if pad and padmode == PAD_PKCS5:
			raise ValueError("Cannot use a pad character with PAD_PKCS5")
		if padmode is None:
			# Get the default padding mode.
			padmode = self.getPadMode()

		if padmode == PAD_NORMAL:
			if not pad:
				# Get the default padding.
				pad = self.getPadding()
			if pad:
				data = data[:-self.block_size] + \
				       data[-self.block_size:].rstrip(pad)

		elif padmode == PAD_PKCS5:
			if _pythonMajorVersion < 3:
				pad_len = ord(data[-1])
			else:
				pad_len = data[-1]
			data = data[:-pad_len]

		return data

	def _guardAgainstUnicode(self, data):
		# Only accept byte strings or ascii unicode values, otherwise
		# there is no way to correctly decode the data into bytes.
		if _pythonMajorVersion < 3:
			if isinstance(data, unicode):
				raise ValueError("pyDes can only work with bytes, not Unicode strings.")
		else:
			if isinstance(data, str):
				# Only accept ascii unicode values.
				try:
					return data.encode('ascii')
				except UnicodeEncodeError:
					pass
				raise ValueError("pyDes can only work with encoded strings, not Unicode.")
		return data

#############################################################################
# 				    DES					    #
#############################################################################
class des(_baseDes):
	"""DES encryption/decrytpion class

	Supports ECB (Electronic Code Book) and CBC (Cypher Block Chaining) modes.

	pyDes.des(key,[mode], [IV])

	key  -> Bytes containing the encryption key, must be exactly 8 bytes
	mode -> Optional argument for encryption type, can be either pyDes.ECB
		(Electronic Code Book), pyDes.CBC (Cypher Block Chaining)
	IV   -> Optional Initial Value bytes, must be supplied if using CBC mode.
		Must be 8 bytes in length.
	pad  -> Optional argument, set the pad character (PAD_NORMAL) to use
		during all encrypt/decrpt operations done with this instance.
	padmode -> Optional argument, set the padding mode (PAD_NORMAL or
		PAD_PKCS5) to use during all encrypt/decrpt operations done
		with this instance.
	"""


	# Permutation and translation tables for DES
	__pc1 = [56, 48, 40, 32, 24, 16,  8,
		  0, 57, 49, 41, 33, 25, 17,
		  9,  1, 58, 50, 42, 34, 26,
		 18, 10,  2, 59, 51, 43, 35,
		 62, 54, 46, 38, 30, 22, 14,
		  6, 61, 53, 45, 37, 29, 21,
		 13,  5, 60, 52, 44, 36, 28,
		 20, 12,  4, 27, 19, 11,  3
	]

	# number left rotations of pc1
	__left_rotations = [
		1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
	]

	# permuted choice key (table 2)
	__pc2 = [
		13, 16, 10, 23,  0,  4,
		 2, 27, 14,  5, 20,  9,
		22, 18, 11,  3, 25,  7,
		15,  6, 26, 19, 12,  1,
		40, 51, 30, 36, 46, 54,
		29, 39, 50, 44, 32, 47,
		43, 48, 38, 55, 33, 52,
		45, 41, 49, 35, 28, 31
	]

	# initial permutation IP
	__ip = [57, 49, 41, 33, 25, 17, 9,  1,
		59, 51, 43, 35, 27, 19, 11, 3,
		61, 53, 45, 37, 29, 21, 13, 5,
		63, 55, 47, 39, 31, 23, 15, 7,
		56, 48, 40, 32, 24, 16, 8,  0,
		58, 50, 42, 34, 26, 18, 10, 2,
		60, 52, 44, 36, 28, 20, 12, 4,
		62, 54, 46, 38, 30, 22, 14, 6
	]

	# Expansion table for turning 32 bit blocks into 48 bits
	__expansion_table = [
		31,  0,  1,  2,  3,  4,
		 3,  4,  5,  6,  7,  8,
		 7,  8,  9, 10, 11, 12,
		11, 12, 13, 14, 15, 16,
		15, 16, 17, 18, 19, 20,
		19, 20, 21, 22, 23, 24,
		23, 24, 25, 26, 27, 28,
		27, 28, 29, 30, 31,  0
	]

	# The (in)famous S-boxes
	__sbox = [
		# S1
		[14, 4, 13, 1, 2, 15, 11, 8, 3, 10, 6, 12, 5, 9, 0, 7,
		 0, 15, 7, 4, 14, 2, 13, 1, 10, 6, 12, 11, 9, 5, 3, 8,
		 4, 1, 14, 8, 13, 6, 2, 11, 15, 12, 9, 7, 3, 10, 5, 0,
		 15, 12, 8, 2, 4, 9, 1, 7, 5, 11, 3, 14, 10, 0, 6, 13],

		# S2
		[15, 1, 8, 14, 6, 11, 3, 4, 9, 7, 2, 13, 12, 0, 5, 10,
		 3, 13, 4, 7, 15, 2, 8, 14, 12, 0, 1, 10, 6, 9, 11, 5,
		 0, 14, 7, 11, 10, 4, 13, 1, 5, 8, 12, 6, 9, 3, 2, 15,
		 13, 8, 10, 1, 3, 15, 4, 2, 11, 6, 7, 12, 0, 5, 14, 9],

		# S3
		[10, 0, 9, 14, 6, 3, 15, 5, 1, 13, 12, 7, 11, 4, 2, 8,
		 13, 7, 0, 9, 3, 4, 6, 10, 2, 8, 5, 14, 12, 11, 15, 1,
		 13, 6, 4, 9, 8, 15, 3, 0, 11, 1, 2, 12, 5, 10, 14, 7,
		 1, 10, 13, 0, 6, 9, 8, 7, 4, 15, 14, 3, 11, 5, 2, 12],

		# S4
		[7, 13, 14, 3, 0, 6, 9, 10, 1, 2, 8, 5, 11, 12, 4, 15,
		 13, 8, 11, 5, 6, 15, 0, 3, 4, 7, 2, 12, 1, 10, 14, 9,
		 10, 6, 9, 0, 12, 11, 7, 13, 15, 1, 3, 14, 5, 2, 8, 4,
		 3, 15, 0, 6, 10, 1, 13, 8, 9, 4, 5, 11, 12, 7, 2, 14],

		# S5
		[2, 12, 4, 1, 7, 10, 11, 6, 8, 5, 3, 15, 13, 0, 14, 9,
		 14, 11, 2, 12, 4, 7, 13, 1, 5, 0, 15, 10, 3, 9, 8, 6,
		 4, 2, 1, 11, 10, 13, 7, 8, 15, 9, 12, 5, 6, 3, 0, 14,
		 11, 8, 12, 7, 1, 14, 2, 13, 6, 15, 0, 9, 10, 4, 5, 3],

		# S6
		[12, 1, 10, 15, 9, 2, 6, 8, 0, 13, 3, 4, 14, 7, 5, 11,
		 10, 15, 4, 2, 7, 12, 9, 5, 6, 1, 13, 14, 0, 11, 3, 8,
		 9, 14, 15, 5, 2, 8, 12, 3, 7, 0, 4, 10, 1, 13, 11, 6,
		 4, 3, 2, 12, 9, 5, 15, 10, 11, 14, 1, 7, 6, 0, 8, 13],

		# S7
		[4, 11, 2, 14, 15, 0, 8, 13, 3, 12, 9, 7, 5, 10, 6, 1,
		 13, 0, 11, 7, 4, 9, 1, 10, 14, 3, 5, 12, 2, 15, 8, 6,
		 1, 4, 11, 13, 12, 3, 7, 14, 10, 15, 6, 8, 0, 5, 9, 2,
		 6, 11, 13, 8, 1, 4, 10, 7, 9, 5, 0, 15, 14, 2, 3, 12],

		# S8
		[13, 2, 8, 4, 6, 15, 11, 1, 10, 9, 3, 14, 5, 0, 12, 7,
		 1, 15, 13, 8, 10, 3, 7, 4, 12, 5, 6, 11, 0, 14, 9, 2,
		 7, 11, 4, 1, 9, 12, 14, 2, 0, 6, 10, 13, 15, 3, 5, 8,
		 2, 1, 14, 7, 4, 10, 8, 13, 15, 12, 9, 0, 3, 5, 6, 11],
	]


	# 32-bit permutation function P used on the output of the S-boxes
	__p = [
		15, 6, 19, 20, 28, 11,
		27, 16, 0, 14, 22, 25,
		4, 17, 30, 9, 1, 7,
		23,13, 31, 26, 2, 8,
		18, 12, 29, 5, 21, 10,
		3, 24
	]

	# final permutation IP^-1
	__fp = [
		39,  7, 47, 15, 55, 23, 63, 31,
		38,  6, 46, 14, 54, 22, 62, 30,
		37,  5, 45, 13, 53, 21, 61, 29,
		36,  4, 44, 12, 52, 20, 60, 28,
		35,  3, 43, 11, 51, 19, 59, 27,
		34,  2, 42, 10, 50, 18, 58, 26,
		33,  1, 41,  9, 49, 17, 57, 25,
		32,  0, 40,  8, 48, 16, 56, 24
	]

	__fpinverse = [
		57, 49, 41, 33, 25, 17, 9,  1,
	    59, 51, 43, 35, 27, 19, 11, 3,
	    61, 53, 45, 37, 29, 21, 13, 5,
	    63, 55, 47, 39, 31, 23, 15, 7,
	    56, 48, 40, 32, 24, 16, 8,  0,
	    58, 50, 42, 34, 26, 18, 10, 2,
	    60, 52, 44, 36, 28, 20, 12, 4,
	    62, 54, 46, 38, 30, 22, 14, 6
	]

	__pinverse = [
		8,  16, 22, 30, 12, 27, 1, 17,
		23, 15, 29, 5,  25, 19, 9, 0,
		7,  13, 24, 2,  3,  28, 10, 18,
		31, 11, 21, 6,  4,  26, 14, 20
	]

	__mykey = [
		0,  2,  3,  4,  5,  1,
		6,  8,  9,  10, 11, 7,
		12, 14, 15, 16, 17, 13,
		18, 20, 21, 22, 23, 19,
		24, 26, 27, 28, 29, 25,
		30, 32, 33, 34, 35, 31,
		36, 38, 39, 40, 41, 37,
		42, 44, 45, 46, 47, 43
		]

	__keylr = [
		4, 23, 6, 15, 5,9,  19, 17, 
		0, 11, 2, 14, 22, 0, 8, 18, 1,
		0, 13, 21, 10, 0, 12, 3, 0, 16, 20, 7,
		46, 30, 26, 47, 34, 40, 0, 45, 27, 0,
		38, 31, 24, 43, 0, 36, 33, 42,  28, 
		35, 37, 44, 32, 25, 41, 0, 29, 39
		]

	


	# Type of crypting being done
	ENCRYPT =	0x00
	DECRYPT =	0x01

	# Initialisation
	def __init__(self, key, mode=ECB, IV=None, pad=None, padmode=PAD_NORMAL):
		# Sanity checking of arguments.
		if len(key) != 64:
			raise ValueError("Invalid DES key size. Key must be exactly 64 bits long.")
		_baseDes.__init__(self, mode, IV, pad, padmode)
		self.key_size = 64

		self.L = []
		self.R = []
		self.Kn = [ [0] * 48 ] * 16	# 16 48-bit keys (K1 - K16)
		self.final = []

		self.setKey(key)

	def setKey(self, key):
		"""Will set the crypting key for this object. Must be 8 bytes."""
		_baseDes.setKey(self, key)
		self.__create_sub_keys()

	

	def __permutate(self, table, block):
		"""Permutate this block with the specified table"""
		return list(map(lambda x: block[x], table))
	
	# Transform the secret key, so that it is ready for data processing
	# Create the 16 subkeys, K[1] - K[16]
	def __create_sub_keys(self):
		"""Create the 16 subkeys K[1] to K[16] from the given key"""
		key = self.__permutate(des.__pc1, self.getKey())
		i = 0
		# Split into Left and Right sections
		self.L = key[:28]
		self.R = key[28:]
		while i < 16:
			j = 0
			# Perform circular left shifts
			while j < des.__left_rotations[i]:
				self.L.append(self.L[0])
				del self.L[0]

				self.R.append(self.R[0])
				del self.R[0]

				j += 1

			# Create one of the 16 subkeys through pc2 permutation
			self.Kn[i] = self.__permutate(des.__pc2, self.L + self.R)	
			i += 1
		
	# Main part of the encryption algorithm, the number cruncher :)
	def __des_crypt(self, block, crypt_type):
		"""Crypt the block of data through DES bit-manipulation"""
		block = self.__permutate(des.__ip, block)
		self.L = block[:32]
		self.R = block[32:]

		# Encryption starts from Kn[1] through to Kn[16]
		if crypt_type == des.ENCRYPT:
			iteration = 0
			iteration_adjustment = 1
		# Decryption starts from Kn[16] down to Kn[1]
		else:
			iteration = 15
			iteration_adjustment = -1

		i = 0
		while i < 16:
			# Make a copy of R[i-1], this will later become L[i]
			tempR = self.R[:]

			# Permutate R[i - 1] to start creating R[i]
			self.R = self.__permutate(des.__expansion_table, self.R)

			# Exclusive or R[i - 1] with K[i], create B[1] to B[8] whilst here
			self.R = list(map(lambda x, y: x ^ y, self.R, self.Kn[iteration]))
			B = [self.R[:6], self.R[6:12], self.R[12:18], self.R[18:24], self.R[24:30], self.R[30:36], self.R[36:42], self.R[42:]]
			# Optimization: Replaced below commented code with above
			#j = 0
			#B = []
			#while j < len(self.R):
			#	self.R[j] = self.R[j] ^ self.Kn[iteration][j]
			#	j += 1
			#	if j % 6 == 0:
			#		B.append(self.R[j-6:j])

			# Permutate B[1] to B[8] using the S-Boxes
			j = 0
			Bn = [0] * 32
			pos = 0
			while j < 8:
				# Work out the offsets
				m = (B[j][0] << 1) + B[j][5]
				n = (B[j][1] << 3) + (B[j][2] << 2) + (B[j][3] << 1) + B[j][4]

				# Find the permutation value
				v = des.__sbox[j][(m << 4) + n]

				# Turn value into bits, add it to result: Bn
				Bn[pos] = (v & 8) >> 3
				Bn[pos + 1] = (v & 4) >> 2
				Bn[pos + 2] = (v & 2) >> 1
				Bn[pos + 3] = v & 1

				pos += 4
				j += 1

			# Permutate the concatination of B[1] to B[8] (Bn)
			self.R = self.__permutate(des.__p, Bn)

			# Xor with L[i - 1]
			self.R = list(map(lambda x, y: x ^ y, self.R, self.L))
			# Optimization: This now replaces the below commented code
			#j = 0
			#while j < len(self.R):
			#	self.R[j] = self.R[j] ^ self.L[j]
			#	j += 1

			# L[i] becomes R[i - 1]
			self.L = tempR

			i += 1
			iteration += iteration_adjustment
		
		# Final permutation of R[16]L[16]
		self.final = self.__permutate(des.__fp, self.R + self.L)

		return self.final


	# Data to be encrypted/decrypted
	def crypt(self, data, crypt_type):
		"""Crypt the data in blocks, running it through des_crypt()"""

		# Error check the data
		if not data:
			return ''
		if len(data) % self.block_size != 0:
			if crypt_type == des.DECRYPT: # Decryption must work on 8 byte blocks
				raise ValueError("Invalid data length, data must be a multiple of " + str(self.block_size) + " bits\n.")
			if not self.getPadding():
				raise ValueError("Invalid data length, data must be a multiple of " + str(self.block_size) + " bits\n. Try setting the optional padding character")
			else:
				data += (self.block_size - (len(data) % self.block_size)) * self.getPadding()
			# print "Len of data: %f" % (len(data) / self.block_size)

		if self.getMode() == CBC:
			if self.getIV():
				#iv = self.__String_to_BitList(self.getIV())
				iv = self.getIV()
			else:
				raise ValueError("For CBC mode, you must supply the Initial Value (IV) for ciphering")

		# Split the data into blocks, crypting each one seperately
		i = 0
		dict = {}
		result = []
		#cached = 0
		#lines = 0
		while i < len(data):
			# Test code for caching encryption results
			#lines += 1
			#if dict.has_key(data[i:i+8]):
				#print "Cached result for: %s" % data[i:i+8]
			#	cached += 1
			#	result.append(dict[data[i:i+8]])
			#	i += 8
			#	continue
				
			#block = self.__String_to_BitList(data[i:i+8])
			block = data[i:i+64]

			# Xor with IV if using CBC mode
			if self.getMode() == CBC:
				if crypt_type == des.ENCRYPT:
					block = list(map(lambda x, y: x ^ y, block, iv))
					#j = 0
					#while j < len(block):
					#	block[j] = block[j] ^ iv[j]
					#	j += 1

				processed_block = self.__des_crypt(block, crypt_type)


				if crypt_type == des.DECRYPT:
					processed_block = list(map(lambda x, y: x ^ y, processed_block, iv))
					#j = 0
					#while j < len(processed_block):
					#	processed_block[j] = processed_block[j] ^ iv[j]
					#	j += 1
					iv = block
				else:
					iv = processed_block
			else:
				processed_block = self.__des_crypt(block, crypt_type)


			# Add the resulting crypted block to our list
			#d = self.__BitList_to_String(processed_block)
			#result.append(d)

			#result.append(self.__BitList_to_String(processed_block))
			
			#dict[data[i:i+8]] = d
			i += 64

		# print "Lines: %d, cached: %d" % (lines, cached)

		# Return the full crypted string
		if _pythonMajorVersion < 3:
			return processed_block
			#return ''.join(result)
		else:
			return processed_block
			#return bytes.fromhex('').join(result)



	def encrypt(self, data, pad=None, padmode=None):
		"""encrypt(data, [pad], [padmode]) -> bytes

		data : Bytes to be encrypted
		pad  : Optional argument for encryption padding. Must only be one byte
		padmode : Optional argument for overriding the padding mode.

		The data must be a multiple of 8 bytes and will be encrypted
		with the already specified key. Data does not have to be a
		multiple of 8 bytes if the padding character is supplied, or
		the padmode is set to PAD_PKCS5, as bytes will then added to
		ensure the be padded data is a multiple of 8 bytes.
		"""
		data = self._padData(data, pad, padmode)
	
		return self.crypt(data, des.ENCRYPT)

	def decrypt(self, data, pad=None, padmode=None):
		"""decrypt(data, [pad], [padmode]) -> bytes

		data : Bytes to be encrypted
		pad  : Optional argument for decryption padding. Must only be one byte
		padmode : Optional argument for overriding the padding mode.

		The data must be a multiple of 8 bytes and will be decrypted
		with the already specified key. In PAD_NORMAL mode, if the
		optional padding character is supplied, then the un-encrypted
		data will have the padding characters removed from the end of
		the bytes. This pad removal only occurs on the last 8 bytes of
		the data (last data block). In PAD_PKCS5 mode, the special
		padding end markers will be removed from the data after decrypting.
		"""
		
		data = self.crypt(data, des.DECRYPT)
		return self._unpadData(data, pad, padmode)

		
	


	def inicio(self, key_in):
		"""Se leen las lineas del texto exportado y se inicia el proceso de obtencion de la CLAVE
		que empieza en leer linea a linea el texto exportado y continua en llamar a las funciones
		oportunas para extraer las claves candidatas para asi almacenarlas en las tablas "mem1" o "mem2". 
		Luego a partir de dichas tablas se encuentran las coincidencias parciales."""

		with open("prueba3.log", "r") as f:
			first_inyection = 1
			for line in f:
				if first_inyection == 1:
					mem1 = self.parse_line(line, first_inyection)
					self.imprime_lista_keys(mem1)
					first_inyection = 0
				else:
					mem2 = self.parse_line(line, first_inyection)
					self.imprime_lista_keys(mem2)
					key_coincide = self.compara_keys(mem1, mem2)
					self.imprime_lista_keys(key_coincide)
					if(key_coincide[0][1]=='U' and key_coincide[1][1]=='U' and key_coincide[2][1]=='U' and
						key_coincide[3][1]=='U' and key_coincide[4][1]=='U' and key_coincide[5][1]=='U' and
						key_coincide[6][1]=='U' and key_coincide[7][1]=='U'):
						break
					else:	
						mem1 = key_coincide
			
			key16 = ['U']*8

			for c in range (8):
				i = conversion()
				key16[c] = key_coincide[c][0]
				key16[c] = i.hex_to_bin(key_coincide[c][0], 6)
			
			key16_list = key16[0] + key16[1]+ key16[2] + key16[3] + key16[4] + key16[5] + key16[6] + key16[7]
			outkey = self.calcula_key_entrada(key16_list, key_in)
			

	def parse_line (self, line, first_inyection):
		"""Se analiza la linea "line" en formato hexadecimal y luego se llama a la funcion
		generator_key pasandole a dicha funcion:outdata_gold y outdata_faulty."""
		
		inj = re.search('([0-9a-fA-F]+):([0-9a-fA-F]+):([0-9a-fA-F]+)', line);
		
		key_candidates = self.generator_key(int(inj.group(2), 16), int(inj.group(3), 16))
		return key_candidates


	def generator_key(self, outdata_gold, outdata_faulty):
		"""A partir de outdata_gold y outdata_faulty se deshace la ultima permutacion y 
		se aplica la diferencia XOR. La mitad derecha es el error inyectado y la mitad
		izquierda es la diferencia XOR de la dos encriptaciones en la salida de la 
		funcion de Feistel. De este modo, a partir de este ultimo valor es posible deshacer
		la permutacion final de Feistel para poder reproducir la ronda 16 para las dos
		encriptaciones y asi poder encontrar coincidencias parciales y almacenarl las
		subclaves de cada caja de sustitucion en una tabla."""

		i= conversion()

		outdata_gold_bits = i.hex_to_bin(outdata_gold, 64)
		outdata_faulty_bits = i.hex_to_bin(outdata_faulty, 64)
		outmsg_gold_bits = self.__permutate(des.__fpinverse, outdata_gold_bits)
		outmsg_faulty_bits = self.__permutate(des.__fpinverse, outdata_faulty_bits)

		
		print("outmsg_gold :   %x"  %i.bin_to_hex(outmsg_gold_bits) )
		print("outmsg_faulty : %x"  %i.bin_to_hex(outmsg_faulty_bits) )

		diferencia = list(map(lambda x, y: x ^ y, outmsg_gold_bits, outmsg_faulty_bits))
		error = diferencia[32:]

		print("error : %x"  %i.bin_to_hex(error))
		
		delta = diferencia[:32]
		p_1 = self.__permutate(des.__pinverse, delta)

		print("inverse permutation : %x"  %i.bin_to_hex(p_1) )

		key = 0
		key_candidates = [['U']*21 for i in range(8)]
		index = [0,0,0,0,0,0,0,0]

		while key <= 63 : 
			
			key_bits = i.hex_to_bin(key, 6)
						
			S_boxes_gold = self.round16(outmsg_gold_bits, key_bits)
			S_boxes_faulty = self.round16(outmsg_faulty_bits, key_bits)
			
			s1_xor_s2 = list(map(lambda x, y: x ^ y, S_boxes_gold, S_boxes_faulty))
			
			indice_error = [31] + list(range(31)) + [0] 
			#Se comprueba si el error inyectado afecta a cada S-Box. Si afecta se comprueban las coincidencias.
			for j in range(8):
			 	if index[j] <= 20:
			 		if (error[indice_error[4*j]]   == 1  or error[indice_error[4*j+5]]  == 1 or 
			 			error[indice_error[4*j+1]] == 1  or error[indice_error[4*j+2]]  == 1 or 
			 			error[indice_error[4*j+3]] == 1  or error[indice_error[4*j+4]]  == 1 ):
			 			if s1_xor_s2[4*j:4*j+4] == p_1[4*j:4*j+4]:
			 				key_candidates[j][index[j]] = key
			 				index[j] += 1
			key += 1
		return key_candidates


	def round16(self, data, key):
		"""Esta funcion emula la ronda 16 a partir de data y key"""
		
		R = self.__permutate(des.__expansion_table, data[32:])
		#print("%x" %i.bin_to_hex(key))	
		key_permutated = [key[0], key[2], key[3], key[4], key[5],key[1]]*8	
		R = list(map(lambda x, y: x ^ y, R, key_permutated))	
		B = [R[:6], R[6:12], R[12:18], R[18:24], R[24:30], R[30:36], R[36:42], R[42:]]

		j = 0
		Bn = [0] * 32
		pos = 0
		while j < 8:
				
			m = (B[j][0] << 1) + B[j][5]
			n = (B[j][1] << 3) + (B[j][2] << 2) + (B[j][3] << 1) + B[j][4]
				# Find the permutation value
			v = des.__sbox[j][(m << 4) + n]
		
				# Turn value into bits, add it to result: Bn
			Bn[pos] = (v & 8) >> 3
			Bn[pos + 1] = (v & 4) >> 2
			Bn[pos + 2] = (v & 2) >> 1
			Bn[pos + 3] = v & 1

			pos += 4
			j += 1
		
		return Bn		


	def imprime_lista_keys (self, lista):
		"""Imprime la lista de subclaves en 8 columnas y en formato hexadecimal"""
		for j  in range(21):
			print()
			for i in range(8):
				if lista[i][j] != 'U':
					if len(hex(lista[i][j]))-2 ==1:
						print ("%x   " %lista[i][j] , end="" )
					elif len(hex(lista[i][j]))-2 ==2:
						print ("%x  " %lista[i][j] , end="" )
				else:
					print("U   ", end="")
		print()
		print()
		print("k1  k2  k3  k4  k5  k6  k7  k8")
		print()

	def compara_keys (self, key1, key2):
		"""Se buscan coincidencias entre key1 y key2 columna por columna"""

		key_coincide = [['U']*21 for i in range(8)]
		indice_coincide = [0]*8

		for i in range(8):
			
			if key1[i][0] == 'U':
				key1[i] = key2[i]
				break
			else:
				if key2[i][0]== 'U':
					key2[i]=key1[i]
					break
				else:
					for j in range(21):		 
						if key1[i][j] != 'U' :
							for k in range(21):		
								if key2[i][k] != 'U':
									if key1[i][j] >= key2[i][k] :
										if key1[i][j] == key2[i][k] :
											key_coincide[i][indice_coincide[i]] = key1[i][j]
											indice_coincide[i] += 1
											break
									else:
										break
								else:
									break
						else:
							break

		return key_coincide




	def calcula_key_entrada(self, key, key_in):
		"""Se deshacen las permutaciones de la clave desde la ronda 16 hasta la 1, y luego 
		se deshace tambien la primera permutacion. Luego se lleva a cabo un ataque de fuerza
		bruta de 8 bits usando como comparacion una encriptacion valida con un texto Palladino
		de prueba y la misma clave que la dada en "tesdes_generator.py"."""


		mykey = self.__permutate(des.__mykey, key)
		keylr = self.__permutate(des.__keylr, mykey)		
				
		keylr[8] = 'U'
		keylr[17] = 'U'
		keylr[21] = 'U'
		keylr[24] = 'U'
		keylr[34] = 'U'
		keylr[37] = 'U'
		keylr[42] = 'U'
		keylr[53] = 'U'

		key_l =[0]*28
		key_r = [0]*28
		key_l[0:27] = keylr[1:28]
		key_l[27] = keylr[0]
		key_r[0:27] = keylr[29:56]
		key_r[27] = keylr[28]

		outkey = [key_l[6], key_l[14], key_l[22], key_r[26], key_r[22], key_r[14], key_r[6],
				 'U', key_l[5]  , key_l[13] , key_l[21], key_r[25], key_r[21], key_r[13], key_r[5],
				 'U', key_l[4], key_l[12], key_l[20], key_r[24], key_r[20], key_r[12], key_r[4],
				 'U', key_l[3], key_l[11], key_l[19], key_r[23], key_r[19], key_r[11],
				 key_r[3], 'U', key_l[2], key_l[10], key_l[18], key_l[26], key_r[18], key_r[10], key_r[2],
				 'U', key_l[1], key_l[9], key_l[17], key_l[25], key_r[17], key_r[9], key_r[1], 'U',
				 key_l[0], key_l[8], key_l[16], key_l[24], key_r[16], key_r[8], key_r[0], 'U', key_l[27],
				 key_l[7], key_l[15], key_l[23], key_r[15], key_r[7], key_r[27], 'U']

		i = conversion()
		keyin = i.hex_to_bin(key_in, 64)
		
		a = des(keyin, CBC, [0]*64, pad=None, padmode=PAD_NORMAL)  
		plaintext = 0x8a74d3ca3b901ff3
		salida_gold = a.encrypt(i.hex_to_bin(plaintext, 64))
		for cont in range(256):
			
			cont_bits = i.hex_to_bin(cont, 8)
			outkey[13] = cont_bits[0]
			outkey[14] = cont_bits[1]
			outkey[18] = cont_bits[2]
			outkey[19] = cont_bits[3]
			outkey[50] = cont_bits[4]
			outkey[53] = cont_bits[5]
			outkey[57] = cont_bits[6]
			outkey[59] = cont_bits[7]
			b = des(outkey, CBC, [0]*64, pad=None, padmode=PAD_NORMAL)  
			salida_barrido = b.encrypt(i.hex_to_bin(plaintext, 64))
			
			if ( salida_gold == salida_barrido ):
				
				print("CLAVE ORIGINAL:")
				print("________________")
				print()
				print(''.join(str(c) for c in i.hex_to_bin(key_in,64)))
				print()
				print("CLAVE OBTENIDA:")
				print("________________")
				print()
				print(''.join(str(c) for c in outkey))
				break
			
class conversion():
	def hex_to_bin(self, data, length):
		
		string_data = list(bin(data))[2:]
		string_data_bits = [int(c) for c in string_data]
		l=len(string_data_bits)
		resultado = [0] * length 
		pos = 0
		i = length - 1
		j = l - 1
		for c in range(l):
			resultado[i]=string_data_bits[j] 
			j -= 1
			i -=1
		return resultado

	def bin_to_hex(self, data):
		
		result = [str(c) for c in data]
		result = ''.join(result)
		result = int(result, 2)	
		return result
